package xyz.matj1.bp.mw_improvements;

import mcdr.test.utils.ObjectCategoryMgmt.Category;
import mcdr.test.utils.ObjectMgmt;
import messif.objects.LocalAbstractObject;
import org.jfree.chart.ChartFactory;
import org.jfree.data.statistics.DefaultBoxAndWhiskerCategoryDataset;
import xyz.matj1.bp.Config;
import xyz.matj1.bp.Utils;
import xyz.matj1.bp.enums.MocapType;
import xyz.matj1.bp.exceptions.CacheIOException;
import xyz.matj1.bp.experiments.KnnSearchTools;
import xyz.matj1.bp.types.MyMap;

import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.List;

import static xyz.matj1.bp.Utils.loadData;
import static xyz.matj1.bp.Utils.readCache;
import static xyz.matj1.bp.Utils.writeCache;

/**
 * This is for experiments with dataset with equalised lengths and motion words with multiple segmentations.
 */
public class EqualLengthExperiments {
	public static void main(String[] args) throws IOException {
		equalisedLengthsPrecisionShowBox(Utils.composeFile(Config.outputDir(), "equalised lengths", "precision box graph.pdf"));
	}

	private static void equalisedLengthsPrecisionShowBox(File output) throws IOException {
		output.getParentFile().mkdirs();
		var knnPrecisions = getKnnPrecisions();
		var dataset = new DefaultBoxAndWhiskerCategoryDataset();

		for (var type : MocapType.values()) {
			var category_List = knnPrecisions.get(type);
			var flatValues = category_List.values()
				.stream()
				.flatMap(Collection::stream)
				.toList();
			dataset.add(flatValues, type, "");
		}

		var graph = ChartFactory.createBoxAndWhiskerChart(
			"",
			"type",
			"precision of kNN search",
			dataset,
			true
		);
		Utils.writeXYGraphAsPdf(
			graph,
			output,
			Config.graphDimensions()
		);
	}

	/**
	 * This gets, in the dataset of sequences with equalised lengths,
	 * for every mocap type, for every category, for every sequence in the category,
	 * the precision of kNN search of the sequence among all other sequences.
	 *
	 * This either computes it or loads it from cache.
	 *
	 * @return a map from every mocap type to a map from every category to a list of all kNN precisions of its sequences
	 * @throws IOException if the data for the computation can't be loaded
	 */
	public static MyMap<MocapType, MyMap<Category, List<Double>>> getKnnPrecisions() throws IOException {
		var knnPrecisions = new MyMap<MocapType, MyMap<Category, List<Double>>>();

		for(var type : MocapType.values()) {
			knnPrecisions.put(type, getKnnPrecisions(type));
		}

		return knnPrecisions;
	}

	/**
	 * This retrieves kNN precisions for every category of sequences of the given mocap type.
	 *
	 * @param type a mocap type on the sequences of which kNN search shall be done
	 * @return a map from every category to a list of precisions, one for each sequence
	 * @throws IOException if the data can't be loaded
	 */
	public static MyMap<Category, List<Double>> getKnnPrecisions(MocapType type) throws IOException {
		var cacheDir = new File(Config.cacheDir(), "eqLenKnnPrecision");
		cacheDir.mkdirs();
		var cachePath = new File(cacheDir, String.format("%s", type.noun()));

		try {
			return readCache(cachePath, MyMap.class);
		} catch(CacheIOException _e) {
			var sequenceMgmt = loadData(
				dataFiles(type),
				type.dataClass()
			);

			var result = KnnSearchTools.computeKnnPrecisions(
				knnSearchResults(type, sequenceMgmt),
				sequenceMgmt
			);

			try {
				writeCache(cachePath, result);
			} catch (CacheIOException e) {
				System.err.printf("Writing cache to %s failed.%n", cachePath);
				e.printStackTrace();
			}

			return result;
		}
	}

	/**
	 * This retrieves kNN search results for every sequence of the given
	 * <code>type</code> among all other sequences of it
	 * by reading cache or by computing the results.
	 *
	 * This loads the dataset from disk. To avoid loading it, if there is one already,
	 * use {@link #knnSearchResults(MocapType, ObjectMgmt)}.
	 *
	 * @param type the type of the sequences for which kNN search shall be done
	 * @return a map from every sequence to a map of its nearest neighbors and their distances
	 */
	public static MyMap<LocalAbstractObject, MyMap<LocalAbstractObject, Float>> knnSearchResults(MocapType type) throws IOException {
		var sequenceMgmt = loadData(dataFiles(type), type.dataClass());
		return knnSearchResults(type, sequenceMgmt);
	}

	/**
	 * This retrieves kNN search results for every sequence of the given
	 * <code>sequenceMgmt</code> among all other sequences of it
	 * by reading cache or by computing the results.
	 *
	 * The given <code>type</code> has to correspond to the type of the sequences in <code>sequenceMgmt</code>.
	 *
	 * This is like {@link #knnSearchResults(MocapType)},
	 * but to avoid loading data if they are already loaded.
	 *
	 * @param type the type of the sequences
	 * @param sequenceMgmt the sequence management with the sequences for kNN
	 * @return a map from every sequence to a map of its nearest neighbors and their distances
	 */
	public static MyMap<LocalAbstractObject, MyMap<LocalAbstractObject, Float>> knnSearchResults(MocapType type, ObjectMgmt sequenceMgmt) {
		var cacheDir = new File(Config.cacheDir(), "eqLenKnnResults");
		return KnnSearchTools.knnSearchResultsWithCaching(type, sequenceMgmt, cacheDir);
	}

	/**
	 * This returns a File pointing to the file with the dataset of the given <code>type</code>.
	 *
	 * @param type the type
	 * @return a File pointing to its dataset
	 */
	public static File dataFiles(MocapType type) {
		return new File(
			Config.dataDir(),
			switch (type) {
				case SKELETON -> "skeleta/class130-actions-coords_normPOS_length-equalised.data";
				case HARD_MW -> "motion words/hard_motion_words_length_equalised.data";
				case SOFT_MW -> "motion words/soft_motion_words_length_equalised.data";
				case MATCH_N_MW -> "motion words/multi-overlay_motion_words_length_equalised.data";
		});
	}
}
